#include "DistortMe.h"

#ifndef MIN
#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#endif

AUDIOCOMPONENT_ENTRY(AUBaseFactory, DistortMePlugin)

void toSample(t_sample *dst, const Float32 *src, long n) {
	while (n--) *dst++ = *src++;
}

void fromSample(Float32 *dst, const t_sample *src, long n) {
	while (n--) *dst++ = *src++;
}

void zeroBuffer(t_sample *dst, int n) {
	while (n--) *dst++ = 0;
}


DistortMePluginKernel::DistortMePluginKernel(DistortMePlugin * inAudioUnit )
:	AUKernelBaseMultichannel(inAudioUnit),
	plug(inAudioUnit),
	bufferSize(512)
{
	inputBuffers.resize(plug->Inputs().GetNumberOfElements());
	outputBuffers.resize(plug->Outputs().GetNumberOfElements());
	
	for(int i=0; i < inputBuffers.size(); i++) {
		inputBuffers[i] = new t_sample[bufferSize];
	}
	for(int i=0; i < outputBuffers.size(); i++) {
		outputBuffers[i] = new t_sample[bufferSize];
	}
}

DistortMePluginKernel::~DistortMePluginKernel() {
	for(int i=0; i < inputBuffers.size(); i++) {
		delete[] inputBuffers[i];
	}
	for(int i=0; i < outputBuffers.size(); i++) {
		delete[] outputBuffers[i];
	}
}

void DistortMePluginKernel::Process(
	const vector<const Float32 *> &inSourceP,
	const vector<Float32 *> &inDestP,
	UInt32 inFramesToProcess,
	bool &ioSilence
) {
	if(!ioSilence) {
		int nin = (int)MIN(inSourceP.size(), inputBuffers.size());
		int nout = (int)MIN(inDestP.size(), outputBuffers.size());
		int n = 0;
		while(n < inFramesToProcess) {
			int toprocess = MIN(bufferSize, (inFramesToProcess-n));
			
			for(int i=0; i < nin; i++) {
				toSample(inputBuffers[i], inSourceP[i]+n, toprocess);
			}
			for(int i=nin; i < inputBuffers.size(); i++) {
				zeroBuffer(inputBuffers[i], toprocess);
			}
			
			perform(
				plug->gen,
				&(inputBuffers[0]), inputBuffers.size(),
				&(outputBuffers[0]), outputBuffers.size(),
				toprocess
			);
			
			for(int i=0; i < nout; i++) {
				fromSample(inDestP[i]+n, outputBuffers[i], toprocess);
			}

			n += toprocess;
		}
	}
}




DistortMePlugin::DistortMePlugin(AudioUnit component)
:	AUEffectBaseMultichannel(component, 2, 1),
	gen(0)
{}

DistortMePlugin::~DistortMePlugin() {
	if(gen) destroy(gen);
}

OSStatus DistortMePlugin::Initialize() {
	OSStatus result = AUEffectBaseMultichannel::Initialize();
	if(result == noErr ) {
	   gen = (CommonState *)create(GetSampleRate(), 512);
		nParams = gen->numparams;
		Globals()->UseIndexedParameters(nParams);
		for(int i=0; i < nParams; i++) {
			ParamInfo *param  = gen->params+i;
			AUEffectBaseMultichannel::SetParameter(i, param->defaultvalue);
		}
	}
	return result;
}


ComponentResult DistortMePlugin::GetParameterInfo(
	AudioUnitScope inScope,
	AudioUnitParameterID inParameterID,
	AudioUnitParameterInfo &outParameterInfo
) {
	ComponentResult result = noErr;
 
    outParameterInfo.flags = kAudioUnitParameterFlag_IsWritable | kAudioUnitParameterFlag_IsReadable;
	
	if (inScope == kAudioUnitScope_Global) {
		if(inParameterID < nParams) {
			ParamInfo *param  = gen->params+inParameterID;
			AUBase::FillInParameterName(
				outParameterInfo,
				CFStringCreateWithCStringNoCopy(NULL, param->name, kCFStringEncodingUTF8, NULL),
				false
			);
			outParameterInfo.minValue = param->outputmin;
			outParameterInfo.maxValue = param->outputmax;
			outParameterInfo.defaultValue = param->defaultvalue;
		}
	}
	else {
		result = kAudioUnitErr_InvalidParameter;
	}
	
	return result;
}

ComponentResult DistortMePlugin::SetParameter(
	AudioUnitParameterID inID,
	AudioUnitScope inScope,
	AudioUnitElement inElement,
	AudioUnitParameterValue inValue,
	UInt32 inBufferOffsetInFrames
) {
	ComponentResult result = noErr;
	
	if (inScope == kAudioUnitScope_Global) {
		if(inID < nParams) {
			::setparameter(gen, inID, inValue, 0);
		}
	}
	else {
		result = kAudioUnitErr_InvalidParameter;
	}
	
	return result;
}